from kuberender import render
from kuberender import file
import click


@click.command()
@click.option('--verbose', '-v', default=False, is_flag=True, help='Whether it should print generated files or not')
@click.option('--context', '-c', 'context_files', help="Yaml file path to be loaded into context. Supports merging.", multiple=True)
@click.option('--set', '-s', 'overriden_vars', help="Vars that override context files. Format: key=value", multiple=True)
@click.option('--template-dir', '-t', default='templates', help='Folder holding templates that should be rendered')
@click.option('--template-url', '-u', default=None, help='URL to download templates from (writes on ~/.kube-render/templates). Accepts URLs on pip format')
@click.option('--apply', '-A', 'should_apply', default=False, is_flag=True, help="Apply rendered files using `kubectl apply`")
@click.option('--working-dir', '-w', default='.', help="Base directory for loading templates and context files")
def cli_render(verbose, template_dir, should_apply, context_files, overriden_vars, template_url, working_dir):
    return_code = render.run(verbose, template_dir, should_apply, context_files, overriden_vars, template_url, working_dir)
    exit(return_code)


@click.command()
@click.option('--verbose', '-v', default=False, is_flag=True, help='Whether it should print generated files or not')
@click.argument('filename')
def cli_file(verbose, filename):
    return_code = file.run(filename, verbose)
    exit(return_code)



from os import path
import sys

from . import render
from .utils import load_yaml_file, merge_dicts


def fix_keys(d):
    return dict((k.replace('-', '_'), v) for (k, v) in d.items())


def configure_working_dir(filename, base_config):
    dirname = path.dirname(filename)
    if not dirname:
        dirname = '.'
    base_config['working-dir'] = dirname
    return base_config


def ensure_apply_by_default(parameters):
    if 'should_apply' not in parameters:
        parameters['should_apply'] = True
    return parameters


def run(filename, verbose=False):
    content = load_yaml_file(filename)
    base_config = content.get('base', {})
    base_config = configure_working_dir(filename, base_config)
    all_renders = content.get('renders', [])
    for r in all_renders:
        parameters = merge_dicts([dict(base_config), r])
        parameters = fix_keys(parameters)
        parameters = ensure_apply_by_default(parameters)
        if verbose:
            parameters['verbose'] = True
            sys.stdout.write('\n\n### RENDERING FIRST ###\n')
            sys.stdout.write(str(parameters) + '\n')
        return_code = render.run(**parameters)
        if return_code != 0:
            return return_code
    return 0



import collections
import os
import subprocess
import sys
from functools import partial
from subprocess import CalledProcessError

import dpath
import jinja2
import yaml
from libvcs.shortcuts import create_repo_from_pip_url

from .utils import load_yaml_file, make_template_path, merge_dicts

RenderedTemplate = collections.namedtuple('RenderedTemplate', ['slug', 'content'])


def should_render_template(template_path):
    filename = template_path.split('/')[-1]
    return not (filename.startswith('.') or filename.startswith('_'))


def render_templates(template_dir, working_dir, **context):
    loader = jinja2.FileSystemLoader([working_dir, template_dir])
    env = jinja2.Environment(loader=loader)
    env.filters['dump'] = lambda o: yaml.dump(o, default_flow_style=False)

    def render(filename):
        template = env.get_template(filename)

        rendered = template.render(yaml=yaml, **context)
        return RenderedTemplate(filename, rendered)
    return [render(t) for t in os.listdir(template_dir) if should_render_template(t)]


def parse_overriden_vars(overriden_vars):
    def parse_statement(override_statement):
        key, value = override_statement.split('=')
        obj = {}
        dpath.util.new(obj, key, value, separator='.')
        return obj

    if isinstance(overriden_vars, dict):
        return [overriden_vars]
    return list(map(parse_statement, overriden_vars))


def update_templates(template_url, dump_dir):
    repo = create_repo_from_pip_url(pip_url=template_url, repo_dir=dump_dir)
    repo.update_repo()


def render(verbose, template_dir, should_apply, context_files, overriden_vars, template_url, working_dir):
    change_working_dir = partial(os.path.join, working_dir)
    context_data = map(load_yaml_file, map(change_working_dir, context_files))

    overriden_vars = parse_overriden_vars(overriden_vars)
    context = merge_dicts(list(context_data) + list(overriden_vars))

    if template_url is not None:
        template_dir = make_template_path(template_url)
        update_templates(template_url, template_dir)
    else:
        template_dir = change_working_dir(template_dir)

    rendered_templates = render_templates(template_dir, working_dir, **context)

    if verbose:
        sys.stdout.write('### Computed context:\n')
        sys.stdout.write(yaml.safe_dump(context, default_flow_style=False, indent=2))
        for t in rendered_templates:
            sys.stdout.write('\n### Rendered {}\n'.format(t.slug))
            sys.stdout.write(t.content)
            sys.stdout.write('\n')
    return rendered_templates


def create_kubectl_apply_pipe():
    return subprocess.Popen(
        ['kubectl', 'apply', '-f', '-'],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.STDOUT
    )


def call_kubectl_apply(template):
    def apply_template(content):
        pipe = create_kubectl_apply_pipe()
        str_content = yaml.safe_dump(content, default_flow_style=False, indent=2)
        output, _ = pipe.communicate(str_content.encode())
        sys.stdout.write(output.decode())
        return_code = abs(pipe.wait())
        if return_code != 0:
            raise CalledProcessError(return_code, pipe.args)
    for t in yaml.load_all(template.content):
        apply_template(t)


def apply_templates(rendered_templates):
    for t in rendered_templates:
        call_kubectl_apply(t)


def run(verbose=False, template_dir='templates', should_apply=False, context_files=None, overriden_vars=None, template_url=None, working_dir='.'):
    context_files = context_files or []
    overriden_vars = overriden_vars or {}
    return_code = 0
    rendered_templates = render(verbose, template_dir, should_apply, context_files, overriden_vars, template_url, working_dir)
    if should_apply:
        try:
            apply_templates(rendered_templates)
        except CalledProcessError as e:
            return_code = e.returncode
    return return_code



from setuptools import setup

__version__ = '0.3.5'

setup(
    name='kuberender',
    version=__version__,
    packages=['kuberender'],
    description='A tool to render and apply Kubernetes manifests using Jinja2',
    install_requires=[
        'Jinja2==2.9.6',
        'click==6.7',
        'PyYAML==3.13',
        'dpath==1.4.0',
        'libvcs==0.3.0',
        'six==1.12.0',
    ],
    entry_points={
        'console_scripts': [
            'kube-render = kuberender:cli_render',
            'kube-render-file = kuberender:cli_file'
        ]
    },
    setup_requires=['pytest-runner'],
    tests_require=['pytest', 'mock']
)



from os.path import expanduser, join
from six.moves.urllib.parse import urlparse


import yaml


def load_yaml_file(path):
    with open(path) as f:
        return yaml.load(f.read())


def merge_dicts(dicts):
    merged = {}
    for d in dicts:
        merged = deep_merge(merged, d, always_concat_list=True)
    return merged


def deep_merge(lhs, rhs, always_concat_list=False):
    def merge_values(k, v_lhs, v_rhs):
        if isinstance(v_lhs, dict) and isinstance(v_rhs, dict):
            return k, deep_merge(v_lhs, v_rhs, always_concat_list)
        elif isinstance(v_lhs, list) and isinstance(v_rhs, list):
            if always_concat_list:
                return k, v_lhs + v_rhs
            else:
                return k, v_rhs
        else:
            return k, v_rhs

    lhs_keys = set(lhs.keys())
    rhs_keys = set(rhs.keys())

    pairs = [merge_values(k, lhs[k], rhs[k]) for k in lhs_keys & rhs_keys] \
        + [(k, lhs[k]) for k in lhs_keys - rhs_keys] \
        + [(k, rhs[k]) for k in rhs_keys - lhs_keys]
    return dict(pairs)


def make_template_path(repo_url):
    parse_result = urlparse(repo_url)
    result_path = parse_result.path
    if parse_result.netloc:
        path = result_path[1:]
    else:
        path = result_path.split(":")[1]
    template_dir = join(expanduser("~"), '.kube-render/templates', path)
    return template_dir
